from .population_engine import PopulationEngine
from .simulation_day import SimulationDay
from .settings import SimfectionSettings
from .logger import SimfectionLogger
from .path import SimfectionPath
from .arguments import _get_parser, simfection_args
import pickle
import time
import os


simfection_logger = SimfectionLogger(name=__name__)
logger = simfection_logger.get_logger()


class SimulationRun():
    def __init__(self, settings: dict = None) -> None:
        logger.info('+ Initializing Simfection Run.')
        # Set settings
        self.settings = SimfectionSettings(settings)
        self.path = SimfectionPath(base_path=self.settings.get_setting('base_path'))

        logger.info('+ Building directory structure at {}.'.format(self.path.base()))
        self.path.build_directory_structure()

        # Needed to avoid numexpr.utils from writing to log
        if 'NUMEXPR_NUM_THREADS' not in os.environ.keys():
            logger.debug('+ Setting NUMEXP_NUM_THREADS to 4.')
            os.environ["NUMEXPR_NUM_THREADS"] = "4"

        if self.settings.get_setting('previous_run') is None:
            self.population = PopulationEngine(self.settings)
            self.population.synthesize_population()
            self.days = None
            self.run_id = 'simfection_{}'.format(int(time.time()))
        else:  # Restarting
            logger.info('+ Restarting from previous run.')
            with open(self.settings.get_setting('previous_run'), 'rb') as _file:
                previous_run = pickle.load(_file)
            self.days = previous_run.days
            self.run_id = previous_run.run_id
            del previous_run
            logger.info('- Loading population.')
            self.population = self.days[-1].population

    def run(self):
        num_days = self.settings.get_setting('num_days')
        population = self.population
        if self.days is None:
            self.days = []
        logger.info('+ Running {} days.'.format(num_days))
        for today in range(num_days):
            # Get day number
            if len(self.days) == 0:
                day_number = today
            else:
                day_number = self.days[-1].day_number + 1
            if today == 0:
                day = SimulationDay(
                    self.run_id,
                    population=population,
                    day_number=day_number,
                    settings=self.settings
                )
            else:
                yesterday = self.days[-1]
                day = SimulationDay(
                    self.run_id,
                    population=yesterday.population,
                    day_number=day_number,
                    settings=self.settings
                )

            day.run()
            self.days.append(day)
            self.path.save_day(day)
        logger.info('- All days ran successfully.')
        logger.info('+ Saving run.')
        self.path.save_run(self)
        self.path.move_log()


def main():
    parser = _get_parser(simfection_args)
    args = parser.parse_args()
    settings = {
        arg: getattr(args, arg)
        for arg in vars(args)
        if getattr(args, arg) is not None
    }
    if settings != {}:
        sim = SimulationRun(settings)
    else:
        sim = SimulationRun()
    sim.run()


if __name__ == '__main__':
    main()



import argparse
from .settings import SimfectionSettings

settings = SimfectionSettings()


def _get_parser(arguments):
    parser = argparse.ArgumentParser()
    for names, args in arguments.items():
        parser.add_argument(*names, **args)
    return parser


simfection_args = {
    ('-nd', '--num-days'): {
        'help': (
            'number of days to simulate \n(default: {})'
            .format(settings.get_setting('num_days'))
        ),
        'type': int,
        'required': False
    },
    ('-bp', '--base-path'): {
        'help': (
            'base path for simfection run \n(default: {})'
            .format(settings.get_setting('base_path'))
        ),
        'type': str,
        'required': False
    },
    ('-pr', '--previous_run'): {
        'help': (
            'path to previous run to use for restart\n(default: {})'
            .format(settings.get_setting('previous_run'))
        ),
        'type': str,
        'required': False
    },
    ('-v', '--verbose'): {
        'help': (
            'whether or not to print extra info to stream \n(default: {})'
            .format(settings.get_setting('verbose'))
        ),
        'type': bool,
        'required': False
    },
    ('-np', '--num-people'): {
        'help': (
            'how many people tp simulation in population \n(default: {})'
            .format(settings.get_setting('num_people'))
        ),
        'type': int,
        'required': False
    },
    ('-if', '--initial_infected'): {
        'help': (
            'ratio of population that is initiall infected \n(default: {})'
            .format(settings.get_setting('initial_states')['inf'])
        ),
        'type': float,
        'required': False
    },
    ('-ir', '--infection-rate'): {
        'help': (
            '\n(default: {})'
            .format(settings.get_setting('infection_rate'))
        ),
        'type': float,
        'required': False
    },
    ('-rr', '--recovery-rate'): {
        'help': (
            '\n(default: {})'
            .format(settings.get_setting('recovery_rate'))
        ),
        'type': float,
        'required': False
    },
    ('-dr', '--death-rate'): {
        'help': (
            '\n(default: {})'
            .format(settings.get_setting('death_rate'))
        ),
        'type': float,
        'required': False
    },
    ('-sr', '--spontaneous-rate'): {
        'help': (
            '\n(default: {})'
            .format(settings.get_setting('spontaneous_rate'))
        ),
        'type': float,
        'required': False
    },
    ('-ip', '--immunity-period'): {
        'help': (
            '\n(default: {})'
            .format(settings.get_setting('immunity_period'))
        ),
        'type': int,
        'required': False
    },
    ('-cr', '--contagious-period'): {
        'help': (
            '\n(default: {})'
            .format(settings.get_setting('contagious_period'))
        ),
        'type': int,
        'required': False
    },
    ('-icp', '--incubation-period'): {
        'help': (
            '\n(default: {})'
            .format(settings.get_setting('incubation_period'))
        ),
        'type': int,
        'required': False
    },
    ('-cpp', '--cpp'): {
        'help': (
            'use the cpp optimization \n(default: {})'
            .format(settings.get_setting('incubation_period'))
        ),
        'dest': 'cpp',
        'action': 'store_true'
    },
}



import logging
import logging.config
import yaml


class SimfectionLogger:
    def __init__(self, name=None):
        # Config and init logger
        # with open('logger.yaml', 'rt') as f:
        #    config = yaml.safe_load(f.read())
        config = {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {'console': {'format': '%(asctime)s  %(name)-30s  %(levelname)s  %(message)s'},
                           'file': {'format': '%(asctime)s %(name)-12s %(levelname)-8s %(message)s'}},
            'handlers': {'console': {'class': 'logging.StreamHandler',
                                     'level': 'INFO',
                                     'formatter': 'console',
                                     'stream': 'ext://sys.stdout'},
                         'file': {'class': 'logging.FileHandler',
                                  'level': 'DEBUG',
                                  'formatter': 'file',
                                  'filename': 'simfection.log',
                                  'encoding': 'utf8'}},
            'root': {'level': 'NOTSET',
                     'handlers': ['console', 'file'],
                     'propagate': True},
            'loggers': {'file': {'handlers': ['file']}}
        }
        logging.config.dictConfig(config)
        if name == '__main__':
            logger = logging.getLogger('simfection')
        else:
            logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)
        self.logger = logger

    def get_logger(self):
        return self.logger



from .connection_engine import ConnectionEngine
from .interaction_engine import InteractionEngine, pathogen
from .population_engine import PopulationEngine
from .settings import SimfectionSettings
from .logger import SimfectionLogger

import pandas as pd
import numpy as np
import time

simfection_logger = SimfectionLogger(name=__name__)
logger = simfection_logger.get_logger()


class UpdateEngine():
    def __init__(self,
                 population: pd.DataFrame = None,
                 settings: SimfectionSettings = None) -> None:
        logger.debug('+ Initializing update engine.')
        self.population = population
        self.verbose = settings.get_setting('verbose')
        logger.debug('- Reading pathogen.')
        self.pathogen = {
            key: settings.get_setting(key) for key in settings.get_setting('pathogen_keys')
        }

    def _update_state(self, agent):

        # Get probabilities
        rec = self.pathogen['recovery_rate']
        dead = self.pathogen['death_rate']
        inf = 1 - (rec + dead)

        # Choose new state
        new_state = np.random.choice(
            ['rec', 'dead', 'inf'],
            p=[rec, dead, inf]
        )

        return new_state

    def _update_states(self):
        population = self.population.copy()

        infected = population.query('state == "inf"').index
        population.loc[infected, 'state'] = population.loc[infected].apply(
            self._update_state,
            axis=1
        )

        self.population = population

    def _update_recovered(self):
        # Get index of those who just recovered
        recovered = self.population.query('state == "rec"').index

        # Reset their state, days_infected, and immunity
        self.population.loc[recovered, 'state'] = 'sus'
        self.population.loc[recovered, 'days_infected'] = 0
        self.population.loc[recovered, 'infected_by'] = None
        self.population.loc[recovered, 'immunity'] = self.pathogen['immunity_period']

    def _update_infected(self):
        # Index of those infected
        infected = self.population.query('state == "inf"').index
        # Add one day infected
        self.population.loc[infected, 'days_infected'] += 1
        # Clear any residual immunity from dummy populations
        self.population.loc[infected, 'immunity'] = 0

    def _update_immune(self):
        # Index of everyone with immunity, except the people who just recovered
        immune = (self.population
                  .query('days_infected > 1')
                  .query('immunity > 0')
                  .index)
        # Remove one day immunity
        self.population.loc[immune, 'immunity'] -= 1

    def update_all(self):
        verbose = self.verbose
        logger.debug('+ Running update engine.')
        logger.debug('- Updating states.')
        self._update_states()
        logger.debug('- Updating recovered.')
        self._update_recovered()
        logger.debug('- Updating infected.')
        self._update_infected()
        logger.debug('- Updating immnue.')
        self._update_immune()
        logger.debug('- Updates complete.')
        if verbose:
            print(self.population.state.value_counts())


if __name__ == '__main__':
    print("Hi, I'm the Connection Engine. I'm not meant to be run directly.")
    print('To use me, please import ConnectionEngine in your script.')



# Class for timing functions
import time
import logging
import typing
import uuid

class Anchor:
    def __init__(self):
        self.start = time.time()
        self.end = None
        self.id = uuid.uuid4()

    def end_anchor(self):
        self.end = time.time()

    def get_start(self) -> float:
        return self.start

    def get_end(self) -> float:
        return self.end

    def timing(self) -> float:
        if(self.end != None):
            return self.end - self.start
        else:
            logging.error("Anchor has no end, cannot return a time.")

class AnchorTracker:

    def __init__(self):
        self.anchors = {}
        self.verbose = False

    def create_anchor(self, anchor_name: str) -> None:
        # Instantiates the anchor
        if anchor_name not in self.anchors:
            self.anchors[anchor_name] = Anchor()
        elif self.verbose:
            logging.warning(f'Anchor \'{anchor_name}\' already exists.')

    def end_anchor(self, anchor_name: str) -> None:
        if anchor_name in self.anchors and self.anchors[anchor_name].get_end() == None:
            # Set the end of the anchor
            self.anchors[anchor_name].end_anchor()
        elif anchor_name in self.anchors and self.anchors[anchor_name].get_end() != None and self.verbose:
            logging.warning(f'Anchor \'{anchor_name}\' already has an end.')
        elif self.verbose:
            logging.warning(f'Anchor \'{anchor_name}\' does not exist.')

    def timing(self, anchor_name: str) -> float:
        if anchor_name in self.anchors and self.anchors[anchor_name].get_end() != None:
            t = self.anchors[anchor_name].get_end() - self.anchors[anchor_name].get_start()
            return t
        elif self.verbose:
            logging.warning(f'Anchor \'{anchor_name}\' does not exist')

    def timing_all(self) -> dict:
        times = {}
        for anchor in self.anchors:
            if self.anchors[anchor].get_end() != None:
                t = self.timing(anchor)
                times[anchor] = t
        return times

    def set_verbose(self, var: bool) -> None:
        self.verbose = var

    def get_anchors(self) -> typing.List[Anchor]:
        return self.anchors.values()

    def __str__(self):
        str = ''
        for anchor in self.anchors:
            str += f'For anchor \'{anchor}\':\n'
            str += f'\tstart_time: {self.anchors[anchor].get_start()}\n'
            if self.anchors[anchor].get_end() != None:
                str += f'\tend_time: {self.anchors[anchor].get_end()}\n'
        return str



import pandas as pd
from .connection_engine import ConnectionEngine
from .settings import SimfectionSettings
from .logger import SimfectionLogger
import numpy as np

simfection_logger = SimfectionLogger(name=__name__)
logger = simfection_logger.get_logger()


class InteractionEngine():
    def __init__(self,
                 connections: pd.DataFrame = None,
                 population: pd.DataFrame = None,
                 settings: SimfectionSettings = None) -> None:
        logger.debug('+ Initializing interaction engine.')
        self.connections = connections
        self.population = population
        self.verbose = settings.get_setting('verbose')
        self.pathogen = {
            key: settings.get_setting(key) for key in settings.get_setting('pathogen_keys')
        }

    def _make_interaction_pair(self, row):
        agent_a = row['agent_a']
        agent_b = row['agent_b']

        if agent_a < agent_b:
            return (row['agent_a'], row['agent_b'])
        else:
            return (row['agent_b'], row['agent_a'])

    def _get_unique_connections(self):
        connections = self.connections

        # Explode on connections
        # Drop num_connections column
        # Rename columns
        connections = (connections.query('num_connections > 0')
                       .explode(column='connections')
                       .drop(columns=['num_connections'])
                       .rename(columns={
                           'agent': 'agent_a',
                           'connections': 'agent_b'
                       })
                       )

        # Combine into interaction pair
        connections['interaction'] = (connections
                                      .apply(self._make_interaction_pair, axis=1)
                                      )

        # Drop unneeded columns
        # Drop duplicates
        connections = (connections
                       .filter(['interaction'])
                       .drop_duplicates()
                       )

        self.connections = connections

    def _get_agents(self, a, b):
        population = self.population

        records = population.query('agent in {}'.format([a, b])).to_dict()
        agent_a = {
            'agent': records['agent'][a],
            'state': records['state'][a]
        }
        agent_b = {
            'agent': records['agent'][b],
            'state': records['state'][b]
        }

        return agent_a, agent_b

    def _qualify_interaction(self, agent_a, agent_b):

        # Determine if agents have different states
        different_states = agent_a['state'] != agent_b['state']

        if different_states:

            # Determine if someone is infected
            someone_infected = agent_a['state'] == 'inf' or agent_b['state'] == 'inf'

            # Determine if someone is susceptible
            someone_susceptible = agent_a['state'] == 'sus' or agent_b['state'] == 'sus'

            if someone_infected and someone_susceptible:

                # Determine who is infected and who is susceptible
                if agent_a['state'] == 'inf':
                    infected = agent_a
                    susceptible = agent_b
                else:
                    infected = agent_b
                    susceptible = agent_a

                # Determine if infected is contagious
                days_infected = (
                    self.population
                    .query('agent == {}'.format(infected['agent']))
                    .days_infected
                    .values[0]
                )
                infected_is_contagious = days_infected < pathogen['contagious_period']

                if infected_is_contagious:

                    # Determine if susecptible is immune
                    immunity = (
                        self.population
                        .query('agent == {}'.format(susceptible['agent']))
                        .immunity
                        .values[0]
                    )
                    susecptible_is_immune = immunity > 0

                    if not susecptible_is_immune:
                        # Interaction qualified
                        return True, infected, susceptible
                    else:
                        # Interaction disqualified
                        return False, None, None
                else:
                    # Interaction disqualified
                    return False, None, None
            else:
                # Interaction disqualified
                return False, None, None
        else:
            # Interaction disqualified
            return False, None, None

    def _interact(self, a, b):
        pathogen = self.pathogen

        # Get agents
        agent_a, agent_b = self._get_agents(a, b)

        # Qualify interaction
        interact, infected, susceptible = self._qualify_interaction(agent_a, agent_b)

        # Interact
        draw = np.random.rand()  # Determines if pathogen spreads
        infect = draw < pathogen['infection_rate']
        if interact and infect:
            self.population.loc[susceptible['agent'], 'state'] = 'inf'
            self.population.loc[susceptible['agent'], 'infected_by'] = (
                int(infected['agent'])
            )

    def interact_all(self):
        verbose = self.verbose
        logger.debug('- Getting unique interactions.')
        self._get_unique_connections()
        connections = self.connections
        logger.debug('- Running all unique interactions.')
        for a, b in connections.interaction.values:
            self._interact(a, b)

        if verbose:
            print(self.population.state.value_counts())


pathogen = {
    'infection_rate': 0.4,
    'recovery_rate': 0.1,
    'death_rate': 0.02,
    'spontaneous_rate': 0.0,
    'testing_accuracy': None,
    'immunity_period': 14,
    'contagious_period': 10,
    'incubation_period': 1,
}

if __name__ == '__main__':
    print("Hi, I'm the Interaction Engine. I'm not meant to be run directly.")
    print('To use me, please import InteractionEngine in your script.')



import os
import shutil
import pickle
import time
from .simulation_day import SimulationDay
from .logger import SimfectionLogger

simfection_logger = SimfectionLogger(name=__name__)
logger = simfection_logger.get_logger()


class SimfectionPath:
    def __init__(self, base_path='simfection'):
        self.base_path = base_path

    # Base directory
    def base(self):
        return self.base_path

    def log(self):
        return os.path.join(self.base(), 'log')

    def output(self):
        return os.path.join(self.base(), 'output')

    def build_directory_structure(self):
        for _dir in [self.base, self.log, self.output]:
            os.makedirs(_dir(), exist_ok=True)

    # Log directory
    def log_file(self, source=False):
        if source:
            return 'simfection.log'
        return os.path.join(self.log(), 'simfection.log')

    def move_log(self):
        source = self.log_file(source=True)
        destination = self.log_file()
        try:
            logger.info('+ Moving log from {} to {}.'.format(source, destination))
            shutil.move(
                source,
                destination
            )
        except FileNotFoundError:
            logger.exception('+ No log file found at {}.'.format(source))

    # Output directory
    def day(self, day):
        day_number = day.day_number
        run_id = day.run_id
        try:
            return os.path.join(self.output(), '{}_day_{}.pickle'.format(run_id, day_number, ))
        except AttributeError:  # For loading when passing in an int
            return os.path.join(self.output(), 'day_{}.pickle'.format(day))

    def save_day(self, day: SimulationDay):
        day_number = day.day_number
        destination = self.day(day)
        logger.debug('+ Saving day {} to {}.'.format(day_number, destination))
        with open(destination, 'wb') as file_:
            pickle.dump(day, file_)

    def load_day(self, day_number: int):
        source = self.day(day_number)
        try:
            logger.info('+ Loading day {} from {}.'.format(day_number, source))
            with open(self.day(day_number), 'rb') as _file:
                return pickle.load(_file)
        except FileNotFoundError:
            logger.exception('+ Unable to load day {} from {}.'.format(day_number, source))

    def save_run(self, run):
        run_id = run.run_id
        destination = os.path.join(self.output(), '{}.pickle'.format(run_id))
        with open(destination, 'wb') as file_:
            pickle.dump(run, file_)



from .connection_engine import ConnectionEngine
from .interaction_engine import InteractionEngine, pathogen
from .population_engine import PopulationEngine
from .update_engine import UpdateEngine
from .settings import SimfectionSettings
from .logger import SimfectionLogger
from .anchor import AnchorTracker

import pandas as pd
import numpy as np
import time
import logging


simfection_logger = SimfectionLogger(name=__name__)
logger = simfection_logger.get_logger()


class SimulationDay():
    def __init__(
            self,
            run_id,
            day_number: int = None,
            population: PopulationEngine = None,
            settings: SimfectionSettings = None) -> None:
        self.anchor_tracker = AnchorTracker()
        assert population is not None or settings is not None, (
            'Both population and settings are NoneType. At least one must be passed.'
        )
        logger.info('+ Simulating day {}.'.format(day_number))
        self.day_number = day_number
        self.settings = settings
        self.run_id = run_id
        if population is None:
            logger.info('+ Dummy population generated.')
            self.population = PopulationEngine(settings)
            self.population.make_dummy()

        else:
            logger.debug('+ Population loaded.')
            self.population = population

        logger.debug('+ Saving starting population.')
        self.starting_population = self.population._df.copy()

    def run(self):
        self.anchor_tracker.create_anchor('run')
        verbose = self.settings.get_setting('verbose')
        self.connection_engine = ConnectionEngine(
            population=self.population._df,
            settings=self.settings
        )
        self.anchor_tracker.create_anchor('connection')
        cpp = self.settings.get_setting('cpp')
        self.connection_engine.create_connections(cpp)
        self.anchor_tracker.end_anchor('connection')
        connection_runtime = self.anchor_tracker.timing('connection')
        logging.debug(f'- Connections made in {connection_runtime} seconds')

        self.interaction_engine = InteractionEngine(
            connections=self.connection_engine.connections,
            settings=self.settings,
            population=self.connection_engine.population
        )
        self.anchor_tracker.create_anchor('interact')
        self.interaction_engine.interact_all()
        self.anchor_tracker.end_anchor('interact')
        interact_runtime = self.anchor_tracker.timing('interact')
        logging.debug(f'- Interactions made in {interact_runtime} seconds.')

        self.update_engine = UpdateEngine(
            population=self.interaction_engine.population,
            settings=self.settings
        )
        self.anchor_tracker.create_anchor('update')
        self.update_engine.update_all()
        self.anchor_tracker.end_anchor('update')
        update_runtime = self.anchor_tracker.timing('update')
        logging.debug(f'- Updates made in {update_runtime} seconds.')

        self.population._df = self.update_engine.population

        self.anchor_tracker.end_anchor('run')
        run_runtime = self.anchor_tracker.timing('run')
        logger.debug('- Day ran successfully.')
        logger.debug(f'- Day simulated in {run_runtime} seconds.')
        logger.debug('- Saving final population.')
        self.final_population = self.population._df



from typing import Any
from .logger import SimfectionLogger

simfection_logger = SimfectionLogger(name=__name__)
logger = simfection_logger.get_logger()


class SimfectionSettings:

    default_settings = {
        # Pathogen settings
        'infection_rate': 0.4,
        'recovery_rate': 0.1,
        'death_rate': 0.00,
        'spontaneous_rate': 0.0,
        'testing_accuracy': None,
        'immunity_period': 10**2,
        'contagious_period': 99,
        'incubation_period': 0,
        'pathogen_keys': [
            'infection_rate',
            'recovery_rate',
            'death_rate',
            'spontaneous_rate',
            'testing_accuracy',
            'immunity_period',
            'contagious_period',
            'incubation_period',
        ],
        # PopulationEngine settings
        'num_people': 100,
        'initial_states': {'inf': 0.2},
        # ConnectionEngine settings
        'mean_connections': 10,
        'std': 10,
        'size': 10**5,
        'experiment': False,
        # SimulationRun settings
        'base_path': 'simfection_run',
        'num_days': 15,
        'verbose': False,
        'previous_run': None,
        # Optimization settings
        'cpp': False
    }

    def __init__(self, settings: dict = None):
        self.settings = self.default_settings
        # Set custom settings
        if settings is not None:
            logger.info('+ Setting custom settings.')
            for key, value in settings.items():
                logger.debug('- Setting {} to {}.'.format(key, value))
                self.set_setting(key, value)

    def get_setting(self, setting: str):
        if setting not in self.settings.keys():
            logger.error('+ Setting "{}" not found.'.format(setting))
            return None
        return self.settings[setting]

    def set_setting(self, setting: str, value: Any):
        if setting not in self.settings.keys():
            logger.error('+ Setting "{}" not found.'.format(setting))
            return None
        self.settings[setting] = value
        return value




import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

requirements = [
    'pandas==1.0.4',
    'PyYAML>=5.0.0',
]
install_requires = [r.strip() for r in requirements]

setuptools.setup(
    name='simfection',
    version='0.1.1',
    author='Simfection Research Group',
    author_email='simfectiongroup@gmail.com',
    description='A stochastic, interaction-based infectious disease model',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/simfection/simfection',
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ],
    python_requires='>=3.7',
    entry_points={
        'console_scripts': [
            'simfection = simfection.simulation_run:main'
        ]
    }
)



import pandas as pd
import numpy as np
from .settings import SimfectionSettings
from .logger import SimfectionLogger

simfection_logger = SimfectionLogger(name=__name__)
logger = simfection_logger.get_logger()


class PopulationEngine:
    def __init__(self,
                 settings: SimfectionSettings,
                 _df: pd.DataFrame = None) -> None:
        logger.debug('+ Initializing population engine.')
        if _df is not None:
            logger.debug('- Loading population.')
            self._df = _df
        else:
            logger.debug('- Unloading settings.')
            self.num_people = settings.get_setting('num_people')
            self.initial_states = settings.get_setting('initial_states')
            self._df = pd.DataFrame()
            self.verbose = settings.get_setting('verbose')

    def _synthesize_states(self):
        initial_states = self.initial_states
        num_people = self.num_people

        # Get total weight
        total_weights = sum([v for k, v in initial_states.items()])
        # Validate given weights
        if total_weights < 1:
            initial_states['sus'] = 1 - total_weights
        elif total_weights > 1:
            raise ValueError(
                '''
                Total weights for initial state cannot exceed 1.
                Total Weight: {}
                Weights: {}
                '''.format(total_weights, initial_states)
            )
        # Synthesize states
        states = np.random.choice(
            [k for k in initial_states.keys()],
            size=num_people,
            p=[v for v in initial_states.values()]
        )

        return states

    def synthesize_population(self):
        logger.info('+ Synthesizing population.')
        num_people = self.num_people
        initial_states = self.initial_states

        population = pd.DataFrame(
            {
                'agent': range(num_people),
                'state': ['sus']*num_people,
                'infected_by': [None]*num_people,
                'days_infected': [0]*num_people,
                'immunity': [0]*num_people
            }
        )

        inf = population.sample(
            n=int(num_people*initial_states['inf'])
        ).agent
        population.loc[inf, 'state'] = 'inf'

        self._df = population

    def make_dummy(self):
        logger.info('+ Synthesizing a dummy population.')
        verbose = self.verbose

        states = ['sus', 'inf', 'dead']

        population = pd.DataFrame(
            {
                'agent': [i for i in range(self.num_people)]
            }
        )
        # Give initial values
        population['state'] = [np.random.choice(states) for i in range(len(population))]
        population['infected_by'] = [[] for i in range(len(population))]
        population['days_infected'] = 0
        population['immunity'] = 0

        # Give susceptible people random immunity
        susceptible = population.query('state == "sus"').index
        population.loc[susceptible, 'immunity'] = [
            np.random.randint(0, high=10)
            for i in range(len(susceptible))
        ]

        # Give infected people random infected_by and days_infected
        infected = population.query('state == "inf"').index
        population.loc[infected, 'days_infected'] = [
            np.random.randint(1, high=10)
            for i in range(len(infected))
        ]
        population.loc[infected, 'infected_by'].apply(
            lambda x: x.append(
                np.random.choice(infected)
            )
        )

        # Give dead people random days_infected and infected_by
        infected = population.query('state == "dead"').index
        population.loc[infected, 'days_infected'] = [
            np.random.randint(0, high=10)
            for i in range(len(infected))
        ]
        population.loc[infected, 'infected_by'].apply(
            lambda x: x.append(
                np.random.choice(infected)
            )
        )

        if verbose:
            print(population.state.value_counts())

        self._df = population



"""Creates a network of hand-shake connections as part of model run.

The ConnectionEngine takes a PopulationEngine object and an average number of connections
and randomly creates a hand-shake network of interactions within the population.

  Typical usage example:

    from settings import SimfectionSettings
    from population_engine import PopulationEngine
    from connection_engine import ConnectionEngine

    # Settings
    settings = SimfectionSettings()

    # Synthesize population
    population = PopulationEngine(settings)
    population.synthesize_population()

    # Create connections
    connection_engine = ConnectionEngine(
      population=population._df,
      settings=settings
    )
    connection_engine.create_connections()
"""
# import typing #TODO: (Grant) I'm not sure if this is
# needed for typing.
import time
import sys
import pandas as pd
import numpy as np
from .settings import SimfectionSettings
from .logger import SimfectionLogger
# the network library depends on the
# network.cpython-37m-darwin.so file being in the same directory
try:
    import network
except ModuleNotFoundError:
    pass

simfection_logger = SimfectionLogger(name=__name__)
logger = simfection_logger.get_logger()

sys.setrecursionlimit(10**6)


class ConnectionEngine():
    """Creates a connections list for each agent in a population following
    the hand-shake protocol.

    Longer class information....
    Longer class information....

    Attributes:
        population: A pandas DataFrame representing a population attribute from
            a PopulationEngine instance.
        mean_connections: An integer average number of connections for each
            agent.
        experiment: A boolean indicating whether or not this instance is being
            used for a timing experiment.
        verbose: A boolean indicating whether or not print debugging
            information to the console.
        std: An integer standard deviation for the distribution from which
            each agent draws to determine the maximum number of connections
            they will have.
        size: An interger number of samples to take from the normal distribution
            from which each agent draws to determine the maximum number of
            connections.
        connections: Initialized as a Nonetype. This will be set in the
            create_connections method.
    """

    def __init__(self, population: pd.DataFrame = None,
                 settings: SimfectionSettings = None) -> None:
        logger.debug('+ Initializing connection engine.')
        self.population = population
        self.mean_connections = settings.get_setting('mean_connections')
        self.experiment = settings.get_setting('experiment')
        self.verbose = settings.get_setting('verbose')
        self.std = settings.get_setting('std')
        self.size = settings.get_setting('size')
        self.connections = None

    def _max_connections(self,
                         std: int = None,
                         size: int = None):
        distribution = np.round(
            np.random.normal(
                loc=self.mean_connections,  # Mean
                scale=std,  # Standard Deviation
                size=size  # sample size
            )
        )

        choice = int(np.random.choice(distribution))
        while choice < 0:
            choice = int(np.random.choice(distribution))

        return choice

    @staticmethod
    def _available_to_connect(agent: str, connections: pd.DataFrame):
        # Return IDs of people with connections less than num_connections
        # Only drop agent if it returns from query
        try:
            return connections[
                connections.num_connections < connections.max_connections
            ].drop(agent).index
        except KeyError:
            return connections[
                connections.num_connections < connections.max_connections
            ].index

    def _build_connection_list(self, agent, connections, use_cpp=False):

        if(use_cpp):
            # Use the wrapped C++ PyConnections object to generate a random network instead
            if self.experiment:
                _start = time.time()
            # Run functions here
            # Instantiate a PyConnections object
            size = connections['agent'].size
            net = network.PyConnections(size)
            # Get the connection limits as a list to pass into function later
            connections_max_list = connections['max_connections'].values.tolist()
            # Generate the random network 2D list to convert to DataFrame form later
            start_gen_random_network = time.time()
            random_network = net.gen_random_network(connections_max_list)
            end_gen_random_network = time.time()
            runtime_gen_random_network = end_gen_random_network - start_gen_random_network
            logger.debug(f'- Random network generated in {runtime_gen_random_network} seconds.')
            # Update the connections DataFrame using the 2D list
            new_connections = pd.DataFrame({'connections': random_network})
            connections.update(new_connections)
            # Update the num_connections part of the connections DataFrame
            # TODO(aogle): update the implementation so it doesn't require magic numbers
            for i in range(0, connections['connections'].size):
                connections.iloc[i, 2] = len(connections.iloc[i, 1])
            if self.experiment:
                runtime_available = time.time() - _start
                runtime_choose = runtime_available
                return connections, runtime_available, runtime_choose
            return connections
        else:
            # Get other agents available to connect
            if self.experiment:
                _start = time.time()
            available = self._available_to_connect(agent, connections)
            if self.experiment:
                runtime_available = time.time() - _start

            # Randomly choose connection
            if self.experiment:
                _start = time.time()
            if len(available) > 0:
                connection = np.random.choice(available)
                # Make connection
                connections.iloc[connection].connections.append(agent)
                connections.iloc[agent].connections.append(connection)

                # Update number of connections
                connections.iloc[[agent, connection], 2] += 1

                # Iterate if necessary
                cont = (
                    connections.num_connections[agent] < connections.max_connections[agent]
                )
                if cont:
                    self._build_connection_list(agent, connections)
            if self.experiment:
                runtime_choose = time.time() - _start
                return connections, runtime_available, runtime_choose
            return connections

    def create_connections(self, use_cpp=False):
        """Creates connection list for each agent.

        A connections DataFrame is created and returned using the agents in the
        PopulationEngine instance self attribute. This method and return object
        spec are required inputs for the next step,which is the interaction
        engine.

        Args:
            ALEX: I don't know if I should put arguments here, because the
            method uses self attributes. Thoughts?

        Returns:
            A pandas DataFrame representing the interaction network for a
            simulation day. Each row is a record that is indexed to an agent in
            the  PopulationEngine instance. A description of the columns
            folowed by an example record are given below.

            DataFrame Columns:
                'agent': unqiue agent ID indexed to the PopulationEngine
                'connections': list of agent IDs with whom the agent had
                    an interaction
                'num_connections': length of list in connections columns
                'max_connections': maximum number of connections the agent can
                    have as randomly drawn

            Example:
                agent connections  num_connections  max_connections
                0         [1]                1                1
                1      [0, 3]                2                1
                2         [5]                1                1
                3      [1, 4]                2                1
                4      [3, 8]                2                1


        Raises:
            #TODO
        """
        std = self.std
        size = self.size
        verbose = self.verbose
        if self.experiment:
            logger.debug('- Entering experiment mode.')

        logger.debug('- Creating connections.')
        not_dead = self.population.query('state != "dead"').index
        connections = pd.DataFrame(
            {
                'agent': list(range(len(not_dead))),
                'connections': [[] for i in not_dead],
                'num_connections': [0 for i in not_dead],
                'max_connections': [
                    self._max_connections(std=std, size=size)
                    for i in not_dead
                ]
            }
        )

        _update = len(not_dead)*0.1
        if self.experiment:
            runtime = {
                'available': [],
                'choose': []
            }
        if(use_cpp):
            # Use the cpp optimization instead
            if self.experiment:
                connections, runtime_available, runtime_choose = self._build_connection_list(
                    0,  # this parameter doesn't matter for cpp
                    connections,
                    use_cpp
                )
                runtime['available'].append(runtime_available)
                runtime['choose'].append(runtime_choose)
            else:
                self._build_connection_list(0, connections, use_cpp)
            self.connections = connections
            logger.debug('- All connections made successfully with C++ optimization.')
            if self.experiment:
                return connections, runtime
            return None
        else:
            for _per in connections.index:

                if self.experiment:
                    connections, runtime_available, runtime_choose = self._build_connection_list(
                        _per,
                        connections,
                    )
                    runtime['available'].append(runtime_available)
                    runtime['choose'].append(runtime_choose)
                else:
                    self._build_connection_list(_per, connections)
            self.connections = connections
            logger.debug('- All connections made successfully.')
            if self.experiment:
                return connections, runtime
            return None


if __name__ == '__main__':
    print("Hi, I'm the Connection Engine. I'm not meant to be run directly.")
    print('To use me, please import ConnectionEngine in your script.')
